/*----------------------------------------------------------------------------*
 *                                                                            *
 *   HPF Array Combining Scatter Function                                     *
 *                                                                            *
 *     - Author: Seungjo Bae                                                  *
 *                                                                            *
 *----------------------------------------------------------------------------*/

#if CM5

  #include <cm/cmmd.h>
  #include <stdio.h>
  #include <math.h>
  #include <stdlib.h>
  
  #include <sys/types.h>
  #include <unistd.h>
  #include <fcntl.h>
  #include <cmsys/cm_fcntl.h>

#else

  #include <mpi.h>
  #include <stdio.h>
  #include <math.h>
  #include <stdlib.h>

#endif

#include "raw_type.h"

#define TAG       3

#define Ceiling(j,k)   ( ( (j) + (k) - 1 ) / (k) )
#define Max(x,y)    ( (x) < (y)   ?   (y)  :  (x) )
#define Min(x,y)    ( (x) < (y)   ?   (x)  :  (y) )

#define Default_local_dense_factor 1.0
#define Default_global_dense_factor 1.0
#define Default_K 4
#define Default_algorithm 0

/***------------------- For debuging ---------------***

typedef enum {false, true} 
    Boolean;

#define SOURCE    0		
#define TARGET    1		
#define master    0		

#if CM5

#else
  extern int only_master;	

#endif

***------------------------------------------------***/

/******************************************************************************
 *                                                                            *
 *  Global Variables for RAW
 *                                                                            *
 ******************************************************************************/

  int PLC;		     /* if PLC == 0, perform partial local combining */

  int Buffer_Unit;	     /* size of Buffer_Type */
  int Data_Unit;	     /* size of Data_Type */

  Data_Types     DATA_TYPE;
  Operator_Types OP_TYPE;

  /*-----------------------------------  MPI  ------------------------------*/

#if CM5
#else
  MPI_Comm  WORLD;
#endif


  /*---------------------------  Distribution  -------------------------------*/

  int tot_nproc;  
  int my_pnum;

  int s_tot_nproc;
  int t_tot_nproc;

/***
  int s_N_all,
      t_N_all;
***/

  /*---------------------------------  Cache  --------------------------------*/

  int *cache_B;


  /*-------------------------  Parameters for Buckets  -----------------------*/

  int K;               /*  K is the number of buckets at each processor  */
  int *count_bucket;   /* holds the number of values to each bucket */
  int *sum_count_bucket;
  int *prefix_sum_count_bucket;
  int *prefix_count_bucket,*bucket_Start;
  int *ptr_bucket;
  
  int  IPsize;		     /* local array size at intermediate processors */
  int  IPsize_all;	     /* total number of elements in dense buckets */

  int  KPS;                /* K * t_tot_nproc */

  int  MP_div_K;	   /* Target bucket size = ceil(t_L_all_max / K) */
  int  virtual_MP;         /* Virtual target local size = MP_div_K * K */
                           /* Used for computing prank */

/***
  double local_dense_factor;
  double global_dense_factor;
***/

  int globally_dense_limit;
  int locally_dense_limit;

  int *Sparse,
      *LSparse;


  /*--------------------  Communication for Dense Buckets  ------------------*/

  int *sendcount;
  int *sdispls;
  int *recvcount;
  int *rdispls;
  int total_send,total_recv;

  /*--- first dtage ---*/
  /*- Buffer_Type *recvbuf,*sendbuf; -*/  /* Make local */

  /*--- second dtage ---*/
  /*- Data_Type *temp; -*/                /* Make local */
  /*- Data_Type *sendbuf2,*recvbuf2; -*/  /* Make local */

  int Size;			/* size of sendbuf2 */


  /*-----------------------------  Segmented Scan  --------------------------*/

  int first_bucket, last_bucket;
  int first_bucket_partial, last_bucket_partial;

  int BN, no_proc, sourcePN;
  int *proc_list;


  /*-------------------  Communication for Sparse Buckets  ------------------*/

  int total_sendSP,total_recvSP;
  int *sendcountSP;
  int *sdisplsSP;
  int *recvcountSP;
  int *rdisplsSP;

  /*- Buffer_Type *recvbufSP,*sendbufSP; -*/    /* Make local */


